@@ -158,6 +158,10 @@ pub struct GroupConfig {
     #[structopt(short = "H", long)]
     pub hard_links: bool,
 
+    /// Don't count matching files found within the same directory argument as duplicates.
+    #[structopt(short("I"), long, conflicts_with("follow-links"))]
+    pub isolate: bool,
+
     /// Before matching, transforms each file by the specified program.
     /// The value of this parameter should contain a command: the path to the program
     /// and optionally a list of space-separated arguments.
@@ -265,10 +269,37 @@ pub struct GroupConfig {
     /// By default descends into directories recursively, unless a recursion depth
     /// limit is specified with `--depth`.
     #[structopt(parse(from_os_str), required_unless("stdin"))]
-    pub paths: Vec<PathBuf>,
+    pub paths: Vec<Path>,
 }
 
 impl GroupConfig {
+    fn validate(&self) -> Result<(), String> {
+        if self.isolate && self.paths.len() <= self.rf_over() {
+            return Err(format!(
+                "The --isolate flag requires that the number of input paths ({}) \
+                 is at least as large as the replication factor lower bound ({}). \
+                 No files would be considered duplicate, regardless of their contents.",
+                self.paths.len(),
+                self.rf_over() + 1,
+            ));
+        }
+        if self.isolate
+            && (self.rf_under.is_some() || self.unique)
+            && self.paths.len() < self.rf_under()
+        {
+            return Err(format!(
+                "The --isolate flag requires that the number of input paths ({}) \
+                 is larger than the replication factor upper bound ({}). \
+                 All files would be considered unique or under-replicated, \
+                 regardless of their contents.",
+                self.paths.len(),
+                self.rf_under() - 1,
+            ));
+        }
+
+        Ok(())
+    }
+
     fn compile_pattern(&self, s: &str) -> Result<Pattern, PatternError> {
         let pattern_opts = if self.caseless {
             PatternOpts::case_insensitive()
@@ -346,13 +377,7 @@ impl GroupConfig {
                     .map(|s| Path::from(s.unwrap().as_str())),
             )
         } else {
-            Box::new(
-                self.paths
-                    .iter()
-                    .map(Path::from)
-                    .collect::<Vec<_>>()
-                    .into_iter(),
-            )
+            Box::new(self.paths.clone().into_iter())
         }
     }
 
@@ -491,7 +516,7 @@ pub enum Command {
     /// Replaces redundant files with links.
     ///
     /// The list of groups earlier produced by `fclones group` should be submitted
-    /// on the standard input. Only the default text format is supported.
+    /// on the standard input.
     ///
     /// Unless `--soft` is specified, hard links are created for links within
     /// the same file system. Soft links are always created to link files between
@@ -505,16 +530,37 @@ pub enum Command {
         soft: bool,
     },
 
+    /// Deduplicates file data using native filesystem deduplication capabilities.
+    ///
+    /// The list of groups earlier produced by `fclones group` should be submitted
+    /// on the standard input.
+    ///
+    /// After successful deduplication, all file clones would still be visible as distinct files,
+    /// but the data would be stored only once, hence taking up possibly less space than before.
+    /// Unlike with hard links, modifying a file does not modify any of its clones.
+    /// The result is not visible to userland applications, so repeated runs
+    /// will find the same files again. This also applies to `fclones dedupe` itself:
+    /// The options `--priority` and `--rf-over` do not detect earlier deduplications.
+    ///
+    /// This command cannot cross file system boundaries.
+    /// Not all file systems support deduplication.
+    /// Not all metadata is preserved on macOS.
+    /// Unsupported on Windows.
+    Dedupe {
+        #[structopt(flatten)]
+        config: DedupeConfig,
+    },
+
     /// Removes redundant files.
     ///
     /// The list of groups earlier produced by `fclones group` should be submitted
-    /// on the standard input. Only the default text format is supported.
+    /// on the standard input.
     Remove(DedupeConfig),
 
     /// Moves redundant files to the given directory.
     ///
     /// The list of groups earlier produced by `fclones group` should be submitted
-    /// on the standard input. Only the default text format is supported.
+    /// on the standard input.
     Move {
         #[structopt(flatten)]
         config: DedupeConfig,
@@ -525,12 +571,32 @@ pub enum Command {
     },
 }
 
+impl Command {
+    pub fn validate(&self) -> Result<(), String> {
+        match self {
+            Command::Group(c) => c.validate(),
+            _ => Ok(()),
+        }
+    }
+}
+
 /// Finds and cleans up redundant files
 #[derive(Debug, StructOpt)]
 #[structopt(
     name = "fclones",
     setting(AppSettings::ColoredHelp),
-    setting(AppSettings::DeriveDisplayOrder)
+    setting(AppSettings::DeriveDisplayOrder),
+    after_help = r#"EXAMPLES:
+
+First call "fclones group -o duplicates.txt directory/" to find all duplicates
+in a directory. See "fclones group --help" on how to modify this search.
+
+Then you may call "fclones SUBCOMMAND < duplicates.txt", where SUBCOMMAND is one
+of link, dedupe, remove or move. See the --help output of each subcommand
+for more information. All have a --dry-run flag to preview changes.
+
+It is also possible to combine the first and second step, e.g.:
+"fclones group dir1/ dir2/ | fclones SUBCOMMAND""#
 )]
 pub struct Config {
     /// Suppresses progress reporting
