@@ -26,7 +26,7 @@ use crate::{Error, FileGroup, TIMESTAMP_FMT};
 use std::collections::HashMap;
 
 /// Defines what to do with redundant files
-#[derive(Clone)]
+#[derive(Clone, PartialEq, Eq)]
 pub enum DedupeOp {
     /// Removes redundant files.
     Remove,
@@ -36,6 +36,8 @@ pub enum DedupeOp {
     SoftLink,
     /// Replaces redundant files with hard-links (ln on Unix).
     HardLink,
+    /// Reflink redundant files (cp --reflink=always, only some filesystems).
+    RefLink,
 }
 
 /// Convenience struct for holding a path to a file and its metadata together
@@ -94,10 +96,14 @@ pub enum FsCommand {
         target: Arc<FileMetadata>,
         link: FileMetadata,
     },
+    RefLink {
+        target: Arc<FileMetadata>,
+        link: FileMetadata,
+    },
 }
 
 impl FsCommand {
-    fn remove(path: &Path) -> io::Result<()> {
+    pub fn remove(path: &Path) -> io::Result<()> {
         let _ = FileLock::new(path)?;
         fs::remove_file(path.to_path_buf())
             .map_err(|e| io::Error::new(e.kind(), format!("Failed to remove file {}: {}", path, e)))
@@ -159,7 +165,7 @@ impl FsCommand {
 
     /// Renames/moves a file from one location to another.
     /// If the target exists, it would be overwritten.
-    fn unsafe_rename(source: &Path, target: &Path) -> io::Result<()> {
+    pub fn unsafe_rename(source: &Path, target: &Path) -> io::Result<()> {
         fs::rename(&source.to_path_buf(), &target.to_path_buf()).map_err(|e| {
             io::Error::new(
                 e.kind(),
@@ -211,7 +217,7 @@ impl FsCommand {
 
     /// Returns a random temporary file name in the same directory, guaranteed to not collide with
     /// any other file in the same directory
-    fn temp_file(path: &Path) -> Path {
+    pub fn temp_file(path: &Path) -> Path {
         let mut name = path
             .file_name()
             .expect("must be a regular file with a name");
@@ -232,7 +238,11 @@ impl FsCommand {
     /// Safely moves the file to a different location and invokes the function.
     /// If the function fails, moves the file back to the original location.
     /// If the function succeeds, removes the file permanently.
-    fn safe_remove<R>(path: &Path, f: impl Fn(&Path) -> io::Result<R>, log: &Log) -> io::Result<R> {
+    pub fn safe_remove<R>(
+        path: &Path,
+        f: impl Fn(&Path) -> io::Result<R>,
+        log: &Log,
+    ) -> io::Result<R> {
         let _ = FileLock::new(path)?; // don't remove a locked file
         let tmp = Self::temp_file(path);
         Self::unsafe_rename(path, &tmp)?;
@@ -271,6 +281,10 @@ impl FsCommand {
                 Self::safe_remove(&link.path, |link| Self::hardlink(&target.path, link), log)?;
                 Ok(FileLen(link.metadata.len()))
             }
+            FsCommand::RefLink { target, link } => {
+                crate::reflink::reflink(target, link, log)?;
+                Ok(FileLen(link.metadata.len()))
+            }
             FsCommand::Move {
                 source,
                 target,
@@ -292,6 +306,7 @@ impl FsCommand {
             FsCommand::Remove { file, .. }
             | FsCommand::SoftLink { link: file, .. }
             | FsCommand::HardLink { link: file, .. }
+            | FsCommand::RefLink { link: file, .. }
             | FsCommand::Move { source: file, .. } => FileLen(file.metadata.len()),
         }
     }
@@ -321,6 +336,15 @@ impl FsCommand {
                 result.push(format!("ln {} {}", target, link));
                 result.push(format!("rm {}", tmp));
             }
+            FsCommand::RefLink { target, link, .. } => {
+                let tmp = Self::temp_file(&link.path);
+                let target = target.path.shell_quote();
+                let link = link.path.shell_quote();
+                // Not really what happens on Linux, there the `mv` is also a reflink.
+                result.push(format!("mv {} {}", link, tmp));
+                result.push(format!("cp --reflink=always {} {}", target, link));
+                result.push(format!("rm {}", tmp));
+            }
             FsCommand::Move {
                 source,
                 target,
@@ -363,6 +387,9 @@ impl FsCommand {
                 result.push(format!("mklink /H {} {}", target, link));
                 result.push(format!("del {}", tmp));
             }
+            FsCommand::RefLink { target, link, .. } => {
+                result.push(format!(":: deduplicate {} {}", link, target));
+            }
             FsCommand::Move {
                 source,
                 target,
@@ -407,8 +434,8 @@ impl AddAssign for DedupeResult {
     }
 }
 
-/// Returns true if any of the files has been modified after given timestamp
-/// Also returns true if file timestamp could
+/// Returns true if any of the files have been modified after the given timestamp.
+/// Also returns true if file timestamp could not be read.
 fn was_modified(files: &[FileMetadata], after: DateTime<FixedOffset>, log: &Log) -> bool {
     let mut result = false;
     let after: DateTime<Local> = after.into();
@@ -592,6 +619,10 @@ impl PartitionedFileGroup {
                     target: retained_file.clone(),
                     link: dropped_file,
                 }),
+                DedupeOp::RefLink => commands.push(FsCommand::RefLink {
+                    target: retained_file.clone(),
+                    link: dropped_file,
+                }),
                 DedupeOp::Remove => commands.push(FsCommand::Remove { file: dropped_file }),
                 DedupeOp::Move(target_dir) => {
                     let source = dropped_file;
