@@ -1,7 +1,7 @@
 use core::fmt;
 use std::cell::RefCell;
 use std::cmp::{max, Reverse};
-use std::collections::HashMap;
+use std::collections::{HashMap, HashSet};
 use std::env::{args, current_dir};
 use std::ffi::{OsStr, OsString};
 use std::fmt::{Display, Formatter};
@@ -48,6 +48,7 @@ mod device;
 mod group;
 mod lock;
 mod pattern;
+mod reflink;
 mod regex;
 mod selector;
 mod semaphore;
@@ -93,6 +94,7 @@ impl From<&str> for Error {
 struct AppCtx<'a> {
     pub config: &'a GroupConfig,
     pub log: &'a Log,
+    input_paths: Vec<Path>,
     devices: DiskDevices,
     transform: Option<Transform>,
     path_selector: PathSelector,
@@ -108,6 +110,12 @@ impl<'a> AppCtx<'a> {
             Some(Err(e)) => return Err(Error::new(format!("Invalid transform: {}", e))),
         };
         let base_dir = Path::from(current_dir().unwrap_or_default());
+        let base_dir_ref = Arc::new(base_dir.clone());
+        let root_paths = config
+            .paths
+            .iter()
+            .map(|p| base_dir_ref.resolve(p))
+            .collect();
         let selector = config
             .path_selector(&base_dir)
             .map_err(|e| format!("Invalid pattern: {}", e))?;
@@ -117,6 +125,7 @@ impl<'a> AppCtx<'a> {
         Ok(AppCtx {
             config,
             log,
+            input_paths: root_paths,
             devices,
             transform,
             path_selector: selector,
@@ -276,13 +285,13 @@ where
 
     let (tx, rx): (Sender<HashedFileInfo>, Receiver<HashedFileInfo>) = channel();
 
-    // There is no point in fclonesing groups containing a single file.
+    // There is no point in processing groups containing a single file.
     // Normally when searching for duplicates such groups are filtered out automatically after
     // each stage, however they are possible when searching for unique files.
     let (groups_to_fclones, groups_to_pass): (Vec<_>, Vec<_>) =
         groups.into_iter().partition(group_pre_filter);
 
-    // This way we can split fclonesing to separate thread-pools, one per device:
+    // This way we can split processing to separate thread-pools, one per device:
     let files = partition_by_devices(groups_to_fclones, devices);
     let mut hash_map =
         GroupMap::new(|f: HashedFileInfo| ((f.file_info.len, f.file_hash), f.file_info));
@@ -299,7 +308,7 @@ where
 
             let tx = tx.clone();
 
-            // Launch a separate thread for each device, so we can fclones
+            // Launch a separate thread for each device, so we can process
             // files on each device independently
             s.spawn(move |_| {
                 // Sort files by their physical location, to reduce disk seek latency
@@ -321,7 +330,7 @@ where
                 // Each task creates a heap allocation and reserves memory in the queue.
                 // It is more memory efficient to keep these tasks as long as possible
                 // in our vector. Without this limit we observed over 20% more memory use
-                // when fclonesing 1M of files.
+                // when processing 1M of files.
                 let semaphore = Arc::new(Semaphore::new(8 * thread_count));
 
                 // Run hashing on the thread-pool dedicated to the device
@@ -332,7 +341,7 @@ where
                     // Spawning a task into a thread-pool requires a static lifetime,
                     // because generally the task could outlive caller's stack frame.
                     // However, this is not the case for rehash function, because
-                    // we don't exit before all tasks are fclonesed.
+                    // we don't exit before all tasks are closed.
                     // In the perfect world we should use scopes for that. Unfortunately
                     // the current implementation of rayon scopes runs the scope body
                     // on one of the thread-pool worker threads, so it is not possible
@@ -421,6 +430,33 @@ fn scan_files(ctx: &AppCtx<'_>) -> Vec<Vec<FileInfo>> {
     files
 }
 
+/// Finds the path root that is a prefix of the given path.
+/// If no match is found, returns the original path.
+fn root_of<'a>(path: &'a Path, roots: &'a [Path]) -> &'a Path {
+    for p in roots {
+        if p.is_prefix_of(path) {
+            return p;
+        }
+    }
+    path
+}
+
+/// Returns `true` if a group of files is big enough to meet the desired grouping criteria.
+/// The number of distinct files in a group must be greater than `rf-over`.
+/// If `--isolate` flag is set, file paths that start with a common input path are counted as one.
+/// Called on each result group after each stage.
+fn should_forward<P: AsPath>(ctx: &AppCtx<'_>, group: &[P]) -> bool {
+    if !ctx.config.isolate {
+        group.len() > ctx.config.rf_over()
+    } else {
+        let s: HashSet<_> = group
+            .iter()
+            .map(|p| root_of(p.path(), &ctx.input_paths))
+            .collect();
+        s.len() > ctx.config.rf_over()
+    }
+}
+
 fn group_by_size(ctx: &AppCtx<'_>, files: Vec<Vec<FileInfo>>) -> Vec<FileGroup<FileInfo>> {
     let file_count: usize = files.iter().map(|v| v.len()).sum();
     let progress = ctx.log.progress_bar("Grouping by size", file_count as u64);
@@ -437,7 +473,7 @@ fn group_by_size(ctx: &AppCtx<'_>, files: Vec<Vec<FileInfo>>) -> Vec<FileGroup<F
 
     let groups: Vec<_> = groups
         .into_iter()
-        .filter(|(_, files)| files.len() > rf_over)
+        .filter(|(_, files)| should_forward(ctx, files.as_slice()))
         .map(|(l, files)| FileGroup {
             file_len: l,
             file_hash: FileHash(0),
@@ -476,6 +512,12 @@ where
                     .unique_by(|p| p.path.hash128()),
             )
         } else {
+            // The BTreeMap inside GroupMap results in a random ordering, so repeated runs
+            // will list a different representative filename first for a given inode.
+            // By ordering these the output becomes deterministic.
+            let mut file_group = file_group;
+            file_group.sort_by(|f1, f2| f1.path.cmp(&f2.path));
+
             files.extend(
                 file_group
                     .into_iter()
@@ -501,7 +543,7 @@ fn remove_same_files(
     let groups: Vec<_> = groups
         .into_par_iter()
         .update(|g| deduplicate(ctx, &mut g.files, |_| progress.tick()))
-        .filter(|g| g.files.len() > rf_over)
+        .filter(|g| should_forward(ctx, g.files.as_slice()))
         .collect();
 
     let count: usize = groups.selected_count(rf_over, rf_under);
@@ -581,7 +623,7 @@ fn group_transformed(
     let groups = rehash(
         groups,
         |_| true,
-        |g| g.files.len() > rf_over,
+        |g| should_forward(ctx, g.files.as_slice()),
         &ctx.devices,
         AccessType::Sequential,
         |(fi, _)| {
@@ -649,7 +691,7 @@ fn group_by_prefix(
     let groups = rehash(
         groups,
         pre_filter,
-        |g| g.files.len() > rf_over,
+        |g| should_forward(ctx, g.files.as_slice()),
         &ctx.devices,
         AccessType::Random,
         |(fi, _)| {
@@ -715,7 +757,7 @@ fn group_by_suffix(ctx: &AppCtx<'_>, groups: Vec<FileGroup<FileInfo>>) -> Vec<Fi
     let groups = rehash(
         groups,
         pre_filter,
-        |g| g.files.len() > rf_over,
+        |g| should_forward(ctx, g.files.as_slice()),
         &ctx.devices,
         AccessType::Random,
         |(fi, old_hash)| {
@@ -761,7 +803,7 @@ fn group_by_contents(
     let groups = rehash(
         groups,
         pre_filter,
-        |g| g.files.len() > rf_over,
+        |g| should_forward(ctx, g.files.as_slice()),
         &ctx.devices,
         AccessType::Sequential,
         |(fi, _)| {
@@ -842,12 +884,12 @@ fn group_by_contents(
 /// ```
 /// use fclones::log::Log;
 /// use fclones::config::GroupConfig;
+/// use fclones::path::Path;
 /// use fclones::{group_files, write_report};
-/// use std::path::PathBuf;
 ///
 /// let log = Log::new();
 /// let mut config = GroupConfig::default();
-/// config.paths = vec![PathBuf::from("/path/to/a/dir")];
+/// config.paths = vec![Path::from("/path/to/a/dir")];
 ///
 /// let groups = group_files(&config, &log).unwrap();
 /// println!("Found {} groups: ", groups.len());
@@ -931,7 +973,7 @@ pub fn write_report(config: &GroupConfig, log: &Log, groups: &[FileGroup<Path>])
 
 #[cfg(test)]
 mod test {
-    use std::fs::{hard_link, File, OpenOptions};
+    use std::fs::{create_dir, hard_link, File, OpenOptions};
     use std::io::{Read, Write};
     use std::path::PathBuf;
     use std::sync::atomic::{AtomicBool, Ordering};
@@ -1137,7 +1179,7 @@ mod test {
 
             let log = test_log();
             let mut config = GroupConfig::default();
-            config.paths = vec![file1, file2];
+            config.paths = vec![file1.into(), file2.into()];
             let results = group_files(&config, &log).unwrap();
             assert_eq!(results.len(), 1);
             assert_eq!(results[0].file_len, FileLen(3));
@@ -1155,7 +1197,7 @@ mod test {
 
             let log = test_log();
             let mut config = GroupConfig::default();
-            config.paths = vec![file1, file2];
+            config.paths = vec![file1.into(), file2.into()];
 
             let results = group_files(&config, &log).unwrap();
             assert_eq!(results.len(), 1);
@@ -1171,6 +1213,9 @@ mod test {
             write_test_file(&file1, b"aaaa", b"", b"");
             write_test_file(&file2, b"aaa", b"", b"");
 
+            let file1 = Path::from(file1);
+            let file2 = Path::from(file2);
+
             let log = test_log();
             let mut config = GroupConfig::default();
             config.paths = vec![file1.clone(), file2.clone()];
@@ -1178,14 +1223,8 @@ mod test {
 
             let results = group_files(&config, &log).unwrap();
             assert_eq!(results.len(), 2);
-            assert_eq!(
-                results[0].files,
-                vec![Path::from(file1.canonicalize().unwrap())]
-            );
-            assert_eq!(
-                results[1].files,
-                vec![Path::from(file2.canonicalize().unwrap())]
-            );
+            assert_eq!(results[0].files, vec![Path::from(file1.canonicalize())]);
+            assert_eq!(results[1].files, vec![Path::from(file2.canonicalize())]);
         });
     }
 
@@ -1199,7 +1238,7 @@ mod test {
 
             let log = test_log();
             let mut config = GroupConfig::default();
-            config.paths = vec![file1.clone(), file2.clone()];
+            config.paths = vec![file1.into(), file2.into()];
             config.unique = true;
 
             let results = group_files(&config, &log).unwrap();
@@ -1221,7 +1260,7 @@ mod test {
 
             let log = test_log();
             let mut config = GroupConfig::default();
-            config.paths = vec![file1.clone(), file2.clone()];
+            config.paths = vec![file1.into(), file2.into()];
             config.unique = true;
 
             let results = group_files(&config, &log).unwrap();
@@ -1243,7 +1282,7 @@ mod test {
 
             let log = test_log();
             let mut config = GroupConfig::default();
-            config.paths = vec![file1.clone(), file2.clone()];
+            config.paths = vec![file1.into(), file2.into()];
             config.unique = true;
 
             let results = group_files(&config, &log).unwrap();
@@ -1263,7 +1302,7 @@ mod test {
 
             let log = test_log();
             let mut config = GroupConfig::default();
-            config.paths = vec![file1.clone(), file2.clone()];
+            config.paths = vec![file1.into(), file2.into()];
             config.unique = true;
 
             let results = group_files(&config, &log).unwrap();
@@ -1279,6 +1318,7 @@ mod test {
             write_test_file(&file1, b"foo", b"", b"");
             let log = test_log();
             let mut config = GroupConfig::default();
+            let file1 = Path::from(file1);
             config.paths = vec![file1.clone(), file1.clone(), file1.clone()];
             config.unique = true;
             config.hard_links = true;
@@ -1304,7 +1344,7 @@ mod test {
 
             let log = test_log();
             let mut config = GroupConfig::default();
-            config.paths = vec![file1.clone(), file2.clone()];
+            config.paths = vec![file1.into(), file2.into()];
             config.unique = true;
             config.hard_links = true;
 
@@ -1314,6 +1354,35 @@ mod test {
         });
     }
 
+    #[test]
+    fn duplicate_files_different_roots() {
+        with_dir("main/duplicate_files_different_roots", |root| {
+            let root1 = root.join("root1");
+            let root2 = root.join("root2");
+            create_dir(&root1).unwrap();
+            create_dir(&root2).unwrap();
+
+            let file1 = root1.join("file1");
+            let file2 = root1.join("file2");
+
+            write_test_file(&file1, b"foo", b"", b"");
+            write_test_file(&file2, b"foo", b"", b"");
+
+            let log = test_log();
+            let mut config = GroupConfig::default();
+            config.paths = vec![root1.into(), root2.into()];
+            config.isolate = true;
+
+            let results = group_files(&config, &log).unwrap();
+            assert_eq!(results.len(), 0);
+
+            config.isolate = false;
+            let results = group_files(&config, &log).unwrap();
+            assert_eq!(results.len(), 1);
+            assert_eq!(results[0].files.len(), 2);
+        });
+    }
+
     #[test]
     fn report() {
         with_dir("main/report", |root| {
@@ -1323,7 +1392,7 @@ mod test {
             let report_file = root.join("report.txt");
             let log = test_log();
             let mut config = GroupConfig::default();
-            config.paths = vec![file.clone()];
+            config.paths = vec![file.into()];
             config.unique = true;
             config.output = Some(report_file.clone());
 
