@@ -12,6 +12,7 @@ use rayon::iter::ParallelBridge;
 use regex::Regex;
 use structopt::StructOpt;
 
+use console::style;
 use fclones::config::{Command, Config, DedupeConfig, GroupConfig, Parallelism};
 use fclones::log::Log;
 use fclones::report::open_report;
@@ -64,21 +65,23 @@ fn run_group(mut config: GroupConfig, log: &mut Log) -> Result<(), Error> {
         // problems and report as warnings.
         let mut access_error = false;
         let depth = config.depth;
-        config.paths.retain(|p| match fs::metadata(&p) {
-            Ok(m) if m.is_dir() && depth == Some(0) => {
-                log.warn(format!(
-                    "Skipping directory {} because recursive scan is disabled.",
-                    p.display()
-                ));
-                false
-            }
-            Err(e) => {
-                log.err(format!("Can't access {}: {}", p.display(), e));
-                access_error = true;
-                false
-            }
-            Ok(_) => true,
-        });
+        config
+            .paths
+            .retain(|p| match fs::metadata(&p.to_path_buf()) {
+                Ok(m) if m.is_dir() && depth == Some(0) => {
+                    log.warn(format!(
+                        "Skipping directory {} because recursive scan is disabled.",
+                        p.display()
+                    ));
+                    false
+                }
+                Err(e) => {
+                    log.err(format!("Can't access {}: {}", p.display(), e));
+                    access_error = true;
+                    false
+                }
+                Ok(_) => true,
+            });
         if access_error {
             return Err(Error::from("Some input paths could not be accessed."));
         }
@@ -177,26 +180,39 @@ pub fn run_dedupe(op: DedupeOp, config: DedupeConfig, log: &mut Log) -> Result<(
         .inspect(|_| progress.tick())
         .par_bridge();
 
+    let upto = if op == DedupeOp::RefLink {
+        // Can't be sure because any previous deduplications are not
+        // visible without calling fs-specific tooling.
+        "up to "
+    } else {
+        ""
+    };
+
     let script = dedupe(groups, op, &dedupe_config, log);
     if dedupe_config.dry_run {
         let out = get_output_writer(&dedupe_config)?;
         let result = log_script(script, out).map_err(|e| format!("Output error: {}", e))?;
         log.info(format!(
-            "Would process {} files and reclaim {} space",
-            result.processed_count, result.reclaimed_space
+            "Would process {} files and reclaim {}{} space",
+            result.processed_count, upto, result.reclaimed_space
         ));
     } else {
         let result = run_script(script, log);
         log.info(format!(
-            "Processed {} files and reclaimed {} space",
-            result.processed_count, result.reclaimed_space
+            "Processed {} files and reclaimed {}{} space",
+            result.processed_count, upto, result.reclaimed_space
         ));
     };
     result.map_err(|e| Error::new(format!("Failed to read file list: {}", e)))
 }
 
 fn main() {
-    let config = Config::from_args();
+    let config: Config = Config::from_args();
+    if let Err(e) = config.command.validate() {
+        eprintln!("{} {}", style("error:").for_stderr().bold().red(), e);
+        exit(1);
+    }
+
     let mut log = Log::new();
     if config.quiet {
         log.no_progress = true;
@@ -218,6 +234,13 @@ fn main() {
             config,
             soft: false,
         } => run_dedupe(DedupeOp::HardLink, config, &mut log),
+        Command::Dedupe { config, .. } => {
+            if cfg!(windows) {
+                log.err("Command \"dedupe\" is unsupported on Windows");
+                exit(1);
+            }
+            run_dedupe(DedupeOp::RefLink, config, &mut log)
+        }
         Command::Move { config, target } => {
             let target = fclones::path::Path::from(target);
             let target = Arc::new(fclones::path::Path::from(cwd)).resolve(target);
