@@ -1,31 +1,29 @@
 //! Output formatting.
 
-use std::io;
-use std::io::{BufRead, BufReader, Error, ErrorKind, Read, Write};
-
-use console::style;
-use lazy_static::lazy_static;
-use regex::Regex;
-use serde::Serialize;
-
-use crate::config::OutputFormat;
-use crate::files::{FileHash, FileLen};
-
-use crate::util::IteratorWrapper;
-use crate::{FileGroup, TIMESTAMP_FMT};
-use chrono::{DateTime, FixedOffset};
-use fallible_iterator::FallibleIterator;
-
 use std::borrow::Borrow;
 use std::cell::Cell;
-
-use crate::path::Path;
 use std::cmp::min;
 use std::fmt::Display;
+use std::io;
+use std::io::{BufRead, BufReader, Error, ErrorKind, Read, Write};
+
+use chrono::{DateTime, FixedOffset};
+use console::style;
+use fallible_iterator::FallibleIterator;
+use itertools::Itertools;
+use lazy_static::lazy_static;
+use regex::Regex;
+use serde::{Deserialize, Serialize};
+
+use crate::config::OutputFormat;
+use crate::files::{FileHash, FileLen};
+use crate::path::Path;
+use crate::util::IteratorWrapper;
+use crate::{FileGroup, TIMESTAMP_FMT};
 
 /// Describes how many redundant files were found, in how many groups,
 /// how much space can be reclaimed, etc.
-#[derive(Debug, Eq, PartialEq, Serialize)]
+#[derive(Clone, Debug, Eq, PartialEq, Serialize, Deserialize)]
 pub struct FileStats {
     pub group_count: usize,
     pub redundant_file_count: usize,
@@ -33,7 +31,7 @@ pub struct FileStats {
 }
 
 /// Data in the header of the whole report.
-#[derive(Debug, Eq, PartialEq, Serialize)]
+#[derive(Clone, Debug, Eq, PartialEq, Serialize, Deserialize)]
 pub struct ReportHeader {
     /// The program version that produced the report
     pub version: String,
@@ -46,6 +44,25 @@ pub struct ReportHeader {
     pub stats: Option<FileStats>,
 }
 
+/// A helper struct that allows to serialize the report with serde.
+/// Together with `IteratorWrapper` used as `groups` it allows to serialize
+/// a report in a streaming way, without the need to keep all groups in memory at once.
+#[derive(Serialize)]
+struct SerializableReport<'a, G: Serialize> {
+    header: &'a ReportHeader,
+    groups: G,
+}
+
+/// A structure for holding contents of the report after fully deserializing the report.
+/// Used only by report readers that deserialize the whole report at once.
+/// Paths are represented as strings, because strings are more memory efficient than Path here,
+/// because we can't do prefix compression that `Path` was designed for.
+#[derive(Deserialize)]
+struct DeserializedReport {
+    header: ReportHeader,
+    groups: Vec<FileGroup<String>>,
+}
+
 /// Formats and writes duplicate files report to a stream.
 /// Supports many formats: text, csv, json, etc.
 pub struct ReportWriter<W: Write> {
@@ -242,13 +259,7 @@ impl<W: Write> ReportWriter<W> {
         I: IntoIterator<Item = P>,
         P: Serialize,
     {
-        #[derive(Serialize)]
-        struct Report<'a, G: Serialize> {
-            header: &'a ReportHeader,
-            groups: G,
-        }
-
-        let report = Report {
+        let report = SerializableReport {
             header,
             groups: IteratorWrapper(Cell::new(Some(groups))),
         };
@@ -260,7 +271,7 @@ impl<W: Write> ReportWriter<W> {
     /// Writes the report in the format given by `format` parameter.
     pub fn write<I, G, P>(
         &mut self,
-        format: &OutputFormat,
+        format: OutputFormat,
         header: &ReportHeader,
         groups: I,
     ) -> io::Result<()>
@@ -278,10 +289,24 @@ impl<W: Write> ReportWriter<W> {
     }
 }
 
+/// Iterator over groups of files, read form the report
+pub type GroupIterator = dyn FallibleIterator<Item = FileGroup<Path>, Error = io::Error> + Send;
+
+/// Reads a report from a stream.
+pub trait ReportReader {
+    /// Reads the header. Must be called exactly once before reading the groups.
+    /// Reports an io::Error with ErrorKind::InvalidData
+    /// if the report header is malformed.
+    fn read_header(&mut self) -> io::Result<ReportHeader>;
+
+    /// Opens an iterator over groups.
+    fn read_groups(self: Box<Self>) -> io::Result<Box<GroupIterator>>;
+}
+
 /// Iterates the contents of the report.
 /// Each emitted item is a group of duplicate files.
-pub struct TextReportIterator<R: Read> {
-    stream: BufReader<R>,
+pub struct TextReportIterator<R: BufRead> {
+    stream: R,
     line_buf: String,
     stopped_on_error: bool,
 }
@@ -296,9 +321,9 @@ struct GroupHeader {
 
 impl<R> TextReportIterator<R>
 where
-    R: Read,
+    R: BufRead,
 {
-    fn new(input: BufReader<R>) -> TextReportIterator<R> {
+    fn new(input: R) -> TextReportIterator<R> {
         TextReportIterator {
             stream: input,
             line_buf: String::new(),
@@ -332,7 +357,7 @@ where
                 Regex::new(r"^([a-f0-9]{32}), ([0-9]+) B [^*]* \* ([0-9]+):").unwrap();
         }
 
-        let captures = GROUP_HEADER_RE.captures(&header_str).ok_or_else(|| {
+        let captures = GROUP_HEADER_RE.captures(header_str).ok_or_else(|| {
             Error::new(
                 ErrorKind::InvalidData,
                 format!("Malformed group header: {}", header_str),
@@ -372,7 +397,7 @@ where
     }
 }
 
-impl<R: Read> FallibleIterator for TextReportIterator<R> {
+impl<R: BufRead + 'static> FallibleIterator for TextReportIterator<R> {
     type Item = FileGroup<Path>;
     type Error = std::io::Error;
 
@@ -403,24 +428,24 @@ impl<R: Read> FallibleIterator for TextReportIterator<R> {
 /// Currently supports only the default text report format.
 /// Does not load the whole report into memory.
 /// Allows iterating over groups of files.
-pub struct TextReportReader<R: Read> {
-    pub header: ReportHeader,
-    pub groups: TextReportIterator<R>,
+pub struct TextReportReader<R: BufRead> {
+    pub stream: R,
 }
 
-impl<R: Read> TextReportReader<R> {
-    fn read_line(stream: &mut impl BufRead) -> io::Result<String> {
+impl<R: BufRead> TextReportReader<R> {
+    /// Creates a new reader for reading from the given stream
+    pub fn new(stream: R) -> TextReportReader<R> {
+        TextReportReader { stream }
+    }
+
+    fn read_line(&mut self) -> io::Result<String> {
         let mut line_buf = String::new();
-        stream.read_line(&mut line_buf)?;
+        self.stream.read_line(&mut line_buf)?;
         Ok(line_buf)
     }
 
-    fn read_extract(
-        stream: &mut impl BufRead,
-        regex: &Regex,
-        msg: &str,
-    ) -> io::Result<Vec<String>> {
-        let line = Self::read_line(stream)?;
+    fn read_extract(&mut self, regex: &Regex, msg: &str) -> io::Result<Vec<String>> {
+        let line = self.read_line()?;
         Ok(regex
             .captures(line.trim())
             .ok_or_else(|| Error::new(ErrorKind::InvalidData, msg.to_owned()))?
@@ -429,14 +454,10 @@ impl<R: Read> TextReportReader<R> {
             .map(|c| c.unwrap().as_str().to_owned())
             .collect())
     }
+}
 
-    /// Creates a new reader for reading from the given stream and
-    /// decodes the report header.
-    /// Reports an io::Error with ErrorKind::InvalidData
-    /// if the report header is malformed.
-    pub fn new(input_stream: R) -> io::Result<TextReportReader<R>> {
-        let mut stream = BufReader::new(input_stream);
-
+impl<R: BufRead + Send + 'static> ReportReader for TextReportReader<R> {
+    fn read_header(&mut self) -> io::Result<ReportHeader> {
         lazy_static! {
             static ref VERSION_RE: Regex =
                 Regex::new(r"^# Report by fclones ([0-9]+\.[0-9]+\.[0-9]+)").unwrap();
@@ -449,54 +470,38 @@ impl<R: Read> TextReportReader<R> {
                     .unwrap();
         }
 
-        let version = Self::read_extract(
-            &mut stream,
-            &VERSION_RE,
-            "Not a default fclones report. \
-            Formats other than the default one are not supported yet.",
-        )?
-        .swap_remove(0);
-        let timestamp = Self::read_extract(
-            &mut stream,
-            &TIMESTAMP_RE,
-            "Malformed header: Missing timestamp",
-        )?
-        .swap_remove(0);
+        let version = self
+            .read_extract(&VERSION_RE, "Malformed header: Missing fclones version")?
+            .swap_remove(0);
+        let timestamp = self
+            .read_extract(&TIMESTAMP_RE, "Malformed header: Missing timestamp")?
+            .swap_remove(0);
         let timestamp = DateTime::parse_from_str(&timestamp, TIMESTAMP_FMT).map_err(|e| {
             Error::new(
                 ErrorKind::InvalidData,
                 format!("Malformed header: Failed to parse timestamp: {}", e),
             )
         })?;
-        let command = Self::read_extract(
-            &mut stream,
-            &COMMAND_RE,
-            "Malformed header: Missing command",
-        )?
-        .swap_remove(0);
+        let command = self
+            .read_extract(&COMMAND_RE, "Malformed header: Missing command")?
+            .swap_remove(0);
         let command = shell_words::split(&command).map_err(|e| {
             Error::new(
                 ErrorKind::InvalidData,
                 format!("Malformed header: Failed to parse command arguments: {}", e),
             )
         })?;
-        let group_count = Self::read_extract(
-            &mut stream,
-            &GROUP_COUNT_RE,
-            "Malformed header: Missing group count",
-        )?
-        .swap_remove(0);
+        let group_count = self
+            .read_extract(&GROUP_COUNT_RE, "Malformed header: Missing group count")?
+            .swap_remove(0);
         let group_count: usize = group_count.parse().map_err(|e| {
             Error::new(
                 ErrorKind::InvalidData,
                 format!("Malformed header: Failed to parse group count: {}", e),
             )
         })?;
-        let stats_line = Self::read_extract(
-            &mut stream,
-            &STATS_RE,
-            "Malformed header: Missing file statistics line",
-        )?;
+        let stats_line =
+            self.read_extract(&STATS_RE, "Malformed header: Missing file statistics line")?;
         let redundant_file_size = FileLen(stats_line[0].parse().map_err(|e| {
             Error::new(
                 ErrorKind::InvalidData,
@@ -516,7 +521,7 @@ impl<R: Read> TextReportReader<R> {
             )
         })?;
 
-        let header = ReportHeader {
+        Ok(ReportHeader {
             version,
             timestamp,
             command,
@@ -525,28 +530,90 @@ impl<R: Read> TextReportReader<R> {
                 redundant_file_count,
                 redundant_file_size,
             }),
-        };
-        Ok(TextReportReader {
-            header,
-            groups: TextReportIterator::new(stream),
         })
     }
+
+    fn read_groups(
+        self: Box<Self>,
+    ) -> io::Result<Box<dyn FallibleIterator<Item = FileGroup<Path>, Error = Error> + Send>> {
+        Ok(Box::new(TextReportIterator::new(self.stream)))
+    }
+}
+
+/// Reads a report from a JSON file.
+/// Currently it is not very memory efficient, because limited to reading the whole file and
+/// deserializing all data into memory.
+pub struct JsonReportReader {
+    report: DeserializedReport,
+}
+
+impl JsonReportReader {
+    pub fn new<R: Read>(stream: R) -> io::Result<JsonReportReader> {
+        let report: DeserializedReport = serde_json::from_reader(stream).map_err(|e| {
+            Error::new(
+                ErrorKind::InvalidData,
+                format!("Failed to deserialize JSON report: {}", e),
+            )
+        })?;
+        Ok(JsonReportReader { report })
+    }
+}
+
+impl ReportReader for JsonReportReader {
+    fn read_header(&mut self) -> io::Result<ReportHeader> {
+        Ok(self.report.header.clone())
+    }
+
+    fn read_groups(self: Box<Self>) -> io::Result<Box<GroupIterator>> {
+        let iter = self.report.groups.into_iter().map(|g| {
+            Ok(FileGroup {
+                file_len: g.file_len,
+                file_hash: g.file_hash,
+                files: g.files.iter().map(|s| Path::from(s.as_str())).collect_vec(),
+            })
+        });
+        let iter = fallible_iterator::convert(iter);
+        Ok(Box::new(iter))
+    }
+}
+
+/// Returns a `ReportReader` that can read and decode the report from the given stream.
+/// Automatically detects the type of the report.
+pub fn open_report(r: impl Read + Send + 'static) -> io::Result<Box<dyn ReportReader>> {
+    let mut buf_reader = BufReader::with_capacity(16 * 1024, r);
+    let preview = buf_reader.fill_buf()?;
+    let preview = String::from_utf8_lossy(preview);
+    if preview.starts_with('{') {
+        Ok(Box::new(JsonReportReader::new(buf_reader)?))
+    } else if preview.starts_with('#') {
+        Ok(Box::new(TextReportReader::new(buf_reader)))
+    } else {
+        Err(io::Error::new(
+            ErrorKind::InvalidData,
+            format!(
+                "Unknown report format. Supported formats are: {}, {}",
+                OutputFormat::Default,
+                OutputFormat::Json
+            ),
+        ))
+    }
 }
 
 #[cfg(test)]
 mod test {
-    use super::*;
+    use tempfile::NamedTempFile;
+
     use crate::files::{FileHash, FileLen};
     use crate::path::Path;
 
-    use chrono::Utc;
-    use tempfile::NamedTempFile;
+    use super::*;
 
     fn dummy_report_header() -> ReportHeader {
         ReportHeader {
             command: vec!["fclones".to_owned(), "find".to_owned(), ".".to_owned()],
             version: env!("CARGO_PKG_VERSION").to_owned(),
-            timestamp: DateTime::from(Utc::now()),
+            timestamp: DateTime::parse_from_str("2021-08-27 12:11:23.456 +0000", TIMESTAMP_FMT)
+                .unwrap(),
             stats: Some(FileStats {
                 group_count: 4,
                 redundant_file_count: 234,
@@ -557,23 +624,21 @@ mod test {
 
     #[test]
     fn test_text_report_reader_reads_header() {
-        let header = dummy_report_header();
+        let header1 = dummy_report_header();
         let groups: Vec<FileGroup<Path>> = vec![];
 
         let output = NamedTempFile::new().unwrap();
         let input = output.reopen().unwrap();
 
         let mut writer = ReportWriter::new(output, false);
-        writer.write_as_text(&header, groups.into_iter()).unwrap();
+        writer.write_as_text(&header1, groups.into_iter()).unwrap();
 
-        let reader = TextReportReader::new(input).unwrap();
-        assert_eq!(reader.header.version, header.version);
-        assert_eq!(reader.header.command, header.command);
-        assert_eq!(
-            reader.header.timestamp.timestamp(),
-            header.timestamp.timestamp()
-        );
-        assert_eq!(reader.header.stats, header.stats);
+        let mut reader = TextReportReader::new(BufReader::new(input));
+        let header2 = reader.read_header().unwrap();
+        assert_eq!(header2.version, header1.version);
+        assert_eq!(header2.command, header1.command);
+        assert_eq!(header2.timestamp.timestamp(), header1.timestamp.timestamp());
+        assert_eq!(header2.stats, header1.stats);
     }
 
     #[test]
@@ -597,9 +662,10 @@ mod test {
 
         let mut writer = ReportWriter::new(output, false);
         writer.write_as_text(&header, groups.iter()).unwrap();
-        let reader = TextReportReader::new(input).unwrap();
+        let mut reader = Box::new(TextReportReader::new(BufReader::new(input)));
+        reader.read_header().unwrap();
 
-        let groups2: Vec<_> = reader.groups.collect().unwrap();
+        let groups2: Vec<_> = reader.read_groups().unwrap().collect().unwrap();
         assert_eq!(groups, groups2);
     }
 
@@ -649,4 +715,70 @@ mod test {
         assert!(g.files.contains(&Path::from("/file3")));
         assert!(g.files.contains(&Path::from("/file4")));
     }
+
+    #[test]
+    fn test_json_report_header() {
+        let header1 = dummy_report_header();
+        let groups: Vec<FileGroup<Path>> = vec![];
+
+        let output = NamedTempFile::new().unwrap();
+        let input = output.reopen().unwrap();
+
+        let mut writer = ReportWriter::new(output, false);
+        writer.write_as_json(&header1, groups.into_iter()).unwrap();
+
+        let mut reader = JsonReportReader::new(input).unwrap();
+        let header2 = reader.read_header().unwrap();
+        assert_eq!(header2.version, header1.version);
+        assert_eq!(header2.command, header1.command);
+        assert_eq!(header2.timestamp.timestamp(), header1.timestamp.timestamp());
+        assert_eq!(header2.stats, header1.stats);
+    }
+
+    #[test]
+    fn test_json_report_reader_reads_files() {
+        let header = dummy_report_header();
+        let groups = vec![
+            FileGroup {
+                file_len: FileLen(100),
+                file_hash: FileHash(0x00112233445566778899aabbccddeeff),
+                files: vec![Path::from("a"), Path::from("b")],
+            },
+            FileGroup {
+                file_len: FileLen(40),
+                file_hash: FileHash(0x0000000000000555555555ffffffffff),
+                files: vec![Path::from("c"), Path::from("d")],
+            },
+        ];
+
+        let output = NamedTempFile::new().unwrap();
+        let input = output.reopen().unwrap();
+
+        let mut writer = ReportWriter::new(output, false);
+        writer.write_as_json(&header, groups.iter()).unwrap();
+        let mut reader = Box::new(JsonReportReader::new(input).unwrap());
+        reader.read_header().unwrap();
+
+        let groups2: Vec<_> = reader.read_groups().unwrap().collect().unwrap();
+        assert_eq!(groups, groups2);
+    }
+
+    fn write_read_header(header: &ReportHeader, format: OutputFormat) -> ReportHeader {
+        let groups: Vec<FileGroup<Path>> = vec![];
+        let output = NamedTempFile::new().unwrap();
+        let input = output.reopen().unwrap();
+        let mut writer = ReportWriter::new(output, false);
+        writer.write(format, header, groups.iter()).unwrap();
+        let mut reader = open_report(input).unwrap();
+        reader.read_header().unwrap()
+    }
+
+    #[test]
+    fn test_format_autodetection() {
+        let header = dummy_report_header();
+        let reread_header_1 = write_read_header(&header, OutputFormat::Default);
+        let reread_header_2 = write_read_header(&header, OutputFormat::Json);
+        assert_eq!(header, reread_header_1);
+        assert_eq!(header, reread_header_2);
+    }
 }
